// filename - SocketHandler.cpp

#include <sys/poll.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <arpa/inet.h>

#include <stdio.h>
#include <strings.h>

#include "SocketHandler.h"

SocketHandler::SocketHandler (int socket_type, int port):
   init_status(Uninitialized),
   port_no(port)
{
   socket_fd = socket(AF_INET, socket_type, 0);
   if (socket_fd < 0)
   {
      perror("Unable to create socket");
   }
}

SocketHandler::~SocketHandler ()
{
   CloseSocket();
}

int SocketHandler::InitializeAsServer ()
{
   if (init_status != Uninitialized)
   {
      perror("Socket already initialized");
      return -1;
   }

   // Declare and clean the address object
   struct sockaddr_in server_addr;
   bzero((char *) &server_addr, sizeof(server_addr));

   // Fill in address and port number information
   server_addr.sin_family = AF_INET;
   server_addr.sin_addr.s_addr = INADDR_ANY; // any ip address of the host
   server_addr.sin_port = htons(port_no);

   auto bind_status = bind(socket_fd, (struct sockaddr *) &server_addr, sizeof(server_addr));
   if (bind_status < 0)
   {
      perror("Socket binding not successful");
   }
   return bind_status;
}

void SocketHandler::ConnectToClient ()
{
   is_client_connected = true;
   if (InitializeAsServer() >= 0)
   {
      puts("Connecting to client");
      listen(socket_fd, 1); // wait for only one connection
      // clean address object and store connection address
      auto size_of_addr = sizeof(connect_to_addr);
      bzero((char *) &connect_to_addr, size_of_addr);
      connected_socket_fd = accept(socket_fd, (struct sockaddr *) &connect_to_addr, (socklen_t *) &size_of_addr);
      close(socket_fd);
      printf("Connection established. Previous socket(%d) replaced by socket(%d)\n", socket_fd, connected_socket_fd);
   }
}

void SocketHandler::ConnectToServer (std::string ip_address)
{
   printf("Connecting to server on port %d. Socket(%d)\n", port_no, socket_fd);
   if (init_status != Uninitialized)
   {
      fputs("Socket already initialized\n", stderr);
   }

   auto client = inet_addr(ip_address.c_str());

   // Declare and clean the address object
   struct sockaddr_in client_addr;
   bzero((char *) &client_addr, sizeof(client_addr));

   // Fill in address and port number information
   client_addr.sin_family = AF_INET;
   client_addr.sin_addr.s_addr = client;
   client_addr.sin_port = htons(port_no);

   if (connect(socket_fd, (struct sockaddr *) &client_addr, sizeof(client_addr)) < 0)
   {
      perror("Cannot connect to server");
   }
}

ssize_t SocketHandler::SendAndWait (const void * data, std::size_t len)
{
   auto retStatus = Send(data, len);
   char recAck;
   Receive(&recAck);
   return retStatus;
}

ssize_t SocketHandler::ReceiveAndAcknowledge (void * buffer, std::size_t len, std::function <void(void *)> PerformBeforeAcknowledging)
{
   struct sockaddr_storage read_from;
   socklen_t read_from_len = sizeof(read_from);
   struct pollfd ds[1];
   ds[0].fd = GetFileDescriptor();
   ds[0].events = POLLIN;
   poll(ds, 1, -1);
   auto retStatus = recvfrom(GetFileDescriptor(), buffer, len, 0, (struct sockaddr *)&read_from, &read_from_len);

   // received data, now perform pre-acknowledgment actionns
   PerformBeforeAcknowledging(buffer);

   auto recAck = 'A';
   sendto(GetFileDescriptor(), &recAck, sizeof(recAck), 0, (struct sockaddr *)&read_from, read_from_len);
   return retStatus;
}

ssize_t SocketHandler::SendLarge (const void * data, std::size_t len)
{
   auto dataOffset = 0;
   while (len > 0)
   {
      auto sentSize = Send(static_cast<const char *>(data) + dataOffset, PACKET_SIZE < len ? PACKET_SIZE : len);
      if (sentSize < 0)
      {
         perror("Message not sent by SocketHandler::SendLarge");
         return -1;
      }
      len -= sentSize;
      dataOffset += sentSize;
   }
   return dataOffset;
}

ssize_t SocketHandler::ReceiveLarge (void * buffer, std::size_t len)
{
   auto dataOffset = 0;
   while (len > 0)
   {
      auto receiveSize = Receive(static_cast<char *>(buffer) + dataOffset, PACKET_SIZE < len ? PACKET_SIZE : len);
      if (receiveSize < 0)
      {
         perror("Message not received by SocketHandler::ReceiveLarge");
         return -1;
      }
      len -= receiveSize;
      dataOffset += receiveSize;
   }
   return dataOffset;
}
