// filename - videoReceiver.cpp

#include "SharedStructs.h"
#include "SignalHandling.h"
#include "SocketHandler.h"

#include <chrono>
#include <condition_variable>
#include <functional>
#include <future>
#include <iostream>
#include <memory>
#include <mutex>
#include <thread>

#include <opencv2/opencv.hpp>

#include "process_metrics.h"

std::mutex tcpReadyLocker;
std::condition_variable tcpReadyEvent;
auto tcpReadyNonSpurious = false;

std::mutex udpReadyLocker;
std::condition_variable udpReadyEvent;
auto udpReadyNonSpurious = false;

// shared resources
std::promise<MatMetaData> * frameMeta = nullptr;
std::function<void ()> ConnectionReset;

void ReceiverMainEntry()
{
   std::cout << "Video receiver ..." << std::endl;

   while (true)
   {
      // TCP server in the main thread
      SocketHandler tcpSocket(SOCK_STREAM, PORT_ONE);
      auto AsyncConnect = [&tcpSocket]{ tcpSocket.ConnectToClient(); return &tcpSocket; };
      auto futureConnection = std::async(std::launch::async, AsyncConnect);

      do
      {
          std::this_thread::sleep_for(std::chrono::milliseconds(200));
      }while (!frameMeta); // while frameMeta is null
      auto frameMetaData = frameMeta->get_future().get();
      frameMeta = nullptr;

      auto dispFrame = cv::Mat{frameMetaData.rows, frameMetaData.cols, frameMetaData.type};
      auto frameSize = dispFrame.total() * dispFrame.elemSize();
      auto framePixels = std::make_unique<uchar[]>(frameSize);
      dispFrame.data = framePixels.get();
//      dispFrame.data = static_cast<uchar *>(operator new(frameSize));

      {
         std::lock_guard<std::mutex> lk(tcpReadyLocker);
           tcpReadyNonSpurious = true;
      }
      tcpReadyEvent.notify_one();

      futureConnection.get(); // ready to get the connection now

      // you can close current socket when it is time to reset the connection
      ConnectionReset = [&tcpSocket]{ tcpSocket.CloseSocket(); };

      cv::namedWindow("Video Receive", cv::WINDOW_AUTOSIZE);

      while (true)
      {
         if (tcpSocket.ReceiveLarge(dispFrame.data, frameSize) >= 0)
         {
            imshow("Video Receive", dispFrame);
         }
         else
         {
            // if connection was terminated, give TCP time to die out
            std::this_thread::sleep_for(std::chrono::seconds(5));
            break; // exit current loop
         }

         if (cv::waitKey(30) == 0) // wait for key press event
         {
            puts("Key press");
            break;
         }
      }
   }
}

void MetaReceiverEntry()
{
   // UDP server in the background thread
   SocketHandler udpSocket(SOCK_DGRAM, PORT_TWO);
   udpSocket.InitializeAsServer();

   auto action = [] (int)
                 {
                    static auto passOnce = true;
                    if (passOnce)
                    {
                       passOnce = false;
                       return;
                    }
                    {
                       std::lock_guard<std::mutex> lk(udpReadyLocker);
                         udpReadyNonSpurious = true;
                    }
                    udpReadyEvent.notify_one();
                 };
   RegisterHandlerForSignal(SIGIO, action, udpSocket.GetFileDescriptor());

   MatMetaData frameInfo;
   while (true)
   {
      std::promise<MatMetaData> p; // a std::promise called p
      frameMeta = &p;
      auto PerformToAcknowledge = [] (void * frameData)
      {
         frameMeta->set_value(*static_cast<MatMetaData *>(frameData));

         std::unique_lock<std::mutex> lk(tcpReadyLocker);
         tcpReadyEvent.wait(lk, []
                                {
                                   auto returnValue = tcpReadyNonSpurious;
                                   tcpReadyNonSpurious = false;
                                   return returnValue; 
                                });
      };

      udpSocket.ReceiveAndAcknowledge(&frameInfo, sizeof(frameInfo), PerformToAcknowledge);

      // wait for another UDP communication from the clinet,
      // then reset the TCP connection
      std::unique_lock<std::mutex> lk(udpReadyLocker);
      udpReadyEvent.wait(lk, []
                             {
                                auto returnValue = udpReadyNonSpurious;
                                udpReadyNonSpurious = false;
                                return returnValue; 
                             });
      ConnectionReset();
   }
}

int main()
{
   std::thread bgthread(MetaReceiverEntry);
   bgthread.detach();
   ReceiverMainEntry();
   return 0;
}
